# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

ap_group = cfg.OptGroup('alarm_processor',
                        title='Alarm processor group',
                        help='Options to configure alarm processor.')

ap_opts = [
    cfg.IntOpt(name='number', min=1, default=2,
               help='Number of alarm processors to spawn',
               deprecated_for_removal=True,
               deprecated_since='1.8.0',
               deprecated_reason='Options is not used in the current code '
                                 'and will be removed in future releases.'),
    cfg.IntOpt(name='ttl', default=14400,
               advanced=True,
               help='Alarms older than TTL are not processed '
                    'by notification engine.')
]

np_group = cfg.OptGroup('notification_processor',
                        title='Notification processor group',
                        help='Options to configure notification processor.')
np_opts = [
    cfg.IntOpt(name='number', min=1,
               default=4, help='Number of notification processors to spawn.')
]


def register_opts(conf):
    conf.register_group(ap_group)
    conf.register_group(np_group)

    conf.register_opts(ap_opts, group=ap_group)
    conf.register_opts(np_opts, group=np_group)


def list_opts():
    return {
        ap_group: ap_opts,
        np_group: np_opts
    }



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from monasca_notification.conf import types

_POSTGRESQL = 'postgresql'
_MYSQL = 'mysql'
_ORM = 'orm'
_REPO_DRIVERS_MAP = {
    _POSTGRESQL: 'monasca_notification.common.repositories.'
                 'postgres.pgsql_repo:PostgresqlRepo',
    _MYSQL: 'monasca_notification.common.repositories.'
            'mysql.mysql_repo:MysqlRepo',
    _ORM: 'monasca_notification.common.repositories.'
          'orm.orm_repo:OrmRepo'
}
_ACCEPTABLE_DRIVER_KEYS = set(list(_REPO_DRIVERS_MAP.keys()) +
                              list(_REPO_DRIVERS_MAP.values()))

_DEFAULT_DB_HOST = '127.0.0.1'
_DEFAULT_DB_USER = 'notification'
_DEFAULT_DB_PASSWORD = 'password'  # nosec bandit B105
_DEFAULT_DB_NAME = 'mon'
_DEFAULT_POSTGRESQL_PORT = 5432
_DEFAULT_MYSQL_PORT = 3306

db_group = cfg.OptGroup('database',
                        title='Database Options',
                        help='Driver configuration for database connectivity.')

db_opts = [
    types.PluginOpt(name='repo_driver', choices=_ACCEPTABLE_DRIVER_KEYS,
                    default=_MYSQL, plugin_map=_REPO_DRIVERS_MAP,
                    required=True, advanced=True,
                    help='Driver name (or full class path) that should be '
                         'used to handle RDB connections. Accepts either '
                         'short labels {0} or full class names {1}. '
                         'Configuring either of them will require presence of '
                         'one of following sections {0} inside configuration '
                         'file.'.format(_REPO_DRIVERS_MAP.keys(),
                                        _REPO_DRIVERS_MAP.values())
                    )
]

orm_group = cfg.OptGroup('orm',
                         title='ORM Options',
                         help='Configuration options to configure '
                              'ORM RBD driver.')
orm_opts = [
    cfg.StrOpt(name='url', default=None,
               help='Connection string for sqlalchemy.')
]

mysql_group = cfg.OptGroup('mysql',
                           title='MySQL Options',
                           help='Configuration options to configure '
                                'plain MySQL RBD driver.')
mysql_opts = [
    cfg.HostAddressOpt(name='host', default=_DEFAULT_DB_HOST,
                       help='IP address of MySQL instance.'),
    cfg.PortOpt(name='port', default=_DEFAULT_MYSQL_PORT,
                help='Port number of MySQL instance.'),
    cfg.StrOpt(name='user', default=_DEFAULT_DB_USER,
               help='Username to connect to MySQL '
                    'instance and given database.'),
    cfg.StrOpt(name='passwd', default=_DEFAULT_DB_PASSWORD,
               ignore_case=True, secret=True,
               help='Password to connect to MySQL instance '
                    'and given database.'),
    cfg.DictOpt(name='ssl', default={},
                help='A dict of arguments similar '
                     'to mysql_ssl_set parameters.'),
    cfg.StrOpt(name='db', default=_DEFAULT_DB_NAME,
               help='Database name available in given MySQL instance.')
]

postgresql_group = cfg.OptGroup('postgresql',
                                title='PostgreSQL Options',
                                help='Configuration options to configure '
                                     'plain PostgreSQL RBD driver.')
postgresql_opts = [
    cfg.HostAddressOpt(name='host', default=_DEFAULT_DB_HOST,
                       help='IP address of PostgreSQL instance.'),
    cfg.PortOpt(name='port', default=_DEFAULT_POSTGRESQL_PORT,
                help='Port number of PostgreSQL instance.'),
    cfg.StrOpt(name='user', default=_DEFAULT_DB_USER,
               help='Username to connect to PostgreSQL '
                    'instance and given database.'),
    cfg.StrOpt(name='password', default=_DEFAULT_DB_PASSWORD,
               secret=True, help='Password to connect to PostgreSQL '
                                 'instance and given database'),
    cfg.StrOpt(name='database', default=_DEFAULT_DB_NAME,
               help='Database name available in '
                    'given PostgreSQL instance.')
]


def register_opts(conf):
    conf.register_group(db_group)
    conf.register_group(orm_group)
    conf.register_group(mysql_group)
    conf.register_group(postgresql_group)

    conf.register_opts(db_opts, group=db_group)
    conf.register_opts(orm_opts, group=orm_group)
    conf.register_opts(mysql_opts, group=mysql_group)
    conf.register_opts(postgresql_opts, group=postgresql_group)


def list_opts():
    return {
        db_group: db_opts,
        orm_group: orm_opts,
        mysql_group: mysql_opts,
        postgresql_group: postgresql_opts,
    }



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

from monasca_common.kafka import client_factory
from monasca_notification.common.utils import construct_notification_object
from monasca_notification.common.utils import get_db_repo
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.processors import notification_processor

log = logging.getLogger(__name__)
CONF = cfg.CONF


class RetryEngine(object):
    def __init__(self):
        self._statsd = get_statsd_client()

        self._consumer = client_factory.get_kafka_consumer(
            CONF.kafka.url,
            CONF.kafka.group,
            CONF.kafka.notification_retry_topic,
            CONF.zookeeper.url,
            CONF.zookeeper.notification_retry_path,
            CONF.kafka.legacy_kafka_client_enabled)
        self._producer = client_factory.get_kafka_producer(
            CONF.kafka.url,
            CONF.kafka.legacy_kafka_client_enabled)

        self._notifier = notification_processor.NotificationProcessor()
        self._db_repo = get_db_repo()

    def run(self):
        for raw_notification in self._consumer:
            message = raw_notification.value()
            notification_data = jsonutils.loads(message)

            notification = construct_notification_object(self._db_repo, notification_data)

            if notification is None:
                self._consumer.commit()
                continue

            wait_duration = CONF.retry_engine.interval - (
                time.time() - notification_data['notification_timestamp'])

            if wait_duration > 0:
                time.sleep(wait_duration)

            sent, failed = self._notifier.send([notification])

            if sent:
                self._producer.publish(CONF.kafka.notification_topic,
                                       [notification.to_json()])

            if failed:
                notification.retry_count += 1
                notification.notification_timestamp = time.time()
                if notification.retry_count < CONF.retry_engine.max_attempts:
                    log.error(u"retry failed for {} with name {} "
                              u"at {}.  "
                              u"Saving for later retry.".format(notification.type,
                                                                notification.name,
                                                                notification.address))
                    self._producer.publish(CONF.kafka.notification_retry_topic,
                                           [notification.to_json()])
                else:
                    log.error(u"retry failed for {} with name {} "
                              u"at {} after {} retries.  "
                              u"Giving up on retry."
                              .format(notification.type,
                                      notification.name,
                                      notification.address,
                                      CONF.retry_engine.max_attempts))

            self._consumer.commit()



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from monasca_notification.conf import types

_DEFAULT_URL = '127.0.0.1:9092'
_DEFAULT_GROUP = 'monasca-notification'
_DEFAULT_ALARM_TOPIC = 'alarm-state-transitions'
_DEFAULT_NOTIFICATION_TOPIC = 'alarm-notifications'
_DEFAULT_RETRY_TOPIC = 'retry-notifications'
_DEFAULT_PERIODIC_TOPICS = {
    '60': '60-seconds-notifications'
}
_DEFAULT_MAX_OFFSET_LAG = 600

kafka_group = cfg.OptGroup('kafka',
                           title='Kafka Options',
                           help='Options under this group allow to configure '
                                'valid connection or Kafka queue.')

kafka_opts = [
    cfg.ListOpt(name='url', item_type=types.HostAddressPortType(),
                bounds=False,
                default=_DEFAULT_URL, required=True,
                help='List of addresses (with ports) pointing '
                     'at Kafka cluster.'),
    cfg.StrOpt(name='group', default=_DEFAULT_GROUP,
               required=True, advanced=True,
               help='Consumer\'s group for monasca-notification client.'),
    cfg.StrOpt(name='alarm_topic', default=_DEFAULT_ALARM_TOPIC,
               required=True, advanced=True,
               help='Topic name in kafka where alarm '
                    'transitions are stored.'),
    cfg.StrOpt(name='notification_topic', default=_DEFAULT_NOTIFICATION_TOPIC,
               required=True, advanced=True,
               help='Topic name in kafka where alarm '
                    'notifications are stored.'),
    cfg.StrOpt(name='notification_retry_topic', default=_DEFAULT_RETRY_TOPIC,
               required=True, advanced=True,
               help='Topic name in kafka where notifications, that have '
                    'failed to be sent and are waiting for retry operations, '
                    'are stored.'),
    cfg.DictOpt(name='periodic', default=_DEFAULT_PERIODIC_TOPICS,
                required=True, advanced=True,
                help='Dict of periodic topics. Keys are the period and '
                     'values the actual topic names in kafka where '
                     'notifications are stored.'),
    cfg.IntOpt(name='max_offset_lag', default=_DEFAULT_MAX_OFFSET_LAG,
               required=True, advanced=True,
               help='Maximum lag for topic that is acceptable by '
                    'the monasca-notification. Notifications that are older '
                    'than this offset are skipped.'),
    cfg.BoolOpt(name='legacy_kafka_client_enabled', default=False,
                required=True, advanced=True,
                help='Enable legacy Kafka client. When set old version of '
                     'kafka-python library is used. Message format version '
                     'for the brokers should be set to 0.9.0.0 to avoid '
                     'performance issues until all consumers are upgraded.')
]


def register_opts(conf):
    conf.register_group(kafka_group)
    conf.register_opts(kafka_opts, group=kafka_group)


def list_opts():
    return {
        kafka_group: kafka_opts
    }



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

retry_engine_group = cfg.OptGroup('retry_engine',
                                  title='Retry Engine Options',
                                  help='Options under this group allow to '
                                       'configure valid connection '
                                       'for retry engine.')

retry_opts = [
    cfg.IntOpt(name='interval', min=30, default=30,
               advanced=True,
               help='How often should retry happen.'),
    cfg.IntOpt(name='max_attempts', default=5,
               advanced=True,
               help='How many times should retrying be tried.')
]


def register_opts(conf):
    conf.register_group(retry_engine_group)
    conf.register_opts(retry_opts, group=retry_engine_group)


def list_opts():
    return {
        retry_engine_group: retry_opts
    }



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import copy

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

from monasca_notification.conf import database
from monasca_notification.conf import kafka
from monasca_notification.conf import keystone
from monasca_notification.conf import notifiers
from monasca_notification.conf import processors
from monasca_notification.conf import queues
from monasca_notification.conf import retry
from monasca_notification.conf import statsd
from monasca_notification.conf import zookeeper
from monasca_notification.plugins import email_notifier
from monasca_notification.plugins import hipchat_notifier
from monasca_notification.plugins import jira_notifier
from monasca_notification.plugins import pagerduty_notifier
from monasca_notification.plugins import slack_notifier
from monasca_notification.plugins import webhook_notifier

LOG = log.getLogger(__name__)
CONF = cfg.CONF

CONF_OPTS = [
    database,
    kafka,
    keystone,
    notifiers,
    processors,
    queues,
    retry,
    statsd,
    zookeeper
]


def register_opts(conf=None):
    if conf is None:
        conf = CONF
    for m in CONF_OPTS:
        m.register_opts(conf)


def register_enabled_plugin_opts(conf=None):
    if conf is None:
        conf = CONF
    for enabled_plugin in conf.notification_types.enabled:
        ep_module = importutils.import_module(enabled_plugin.split(":")[0])
        ep_module.register_opts(conf)


def list_opts():
    PLUGIN_CONF_OPTS = [slack_notifier, jira_notifier,
                        hipchat_notifier, email_notifier,
                        webhook_notifier, pagerduty_notifier]
    CONF_OPTS.extend(PLUGIN_CONF_OPTS)
    opts = collections.defaultdict(list)
    for m in CONF_OPTS:
        configs = copy.deepcopy(m.list_opts())
        for key, val in configs.items():
            opts[key].extend(val)
    return _tupleize(opts)


def _tupleize(d):
    """Convert a dict of options to the 2-tuple format."""
    return [(key, value) for key, value in d.items()]



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

_DEFAULT_HOST = '127.0.0.1'
_DEFAULT_PORT = 8125

statsd_group = cfg.OptGroup('statsd',
                            title='statsd Options',
                            help='Options under this group allow '
                            'to configure valid connection '
                            'to statsd server launched by monasca-agent.')

statsd_opts = [
    cfg.BoolOpt('enable', default=True,
                help='Enable or disable self monitoring.'),
    cfg.HostAddressOpt('host', default=_DEFAULT_HOST,
                       help='IP address of statsd server.'),
    cfg.PortOpt('port', default=_DEFAULT_PORT, help='Port of statsd server.'),
]


def register_opts(conf):
    conf.register_group(statsd_group)
    conf.register_opts(statsd_opts, group=statsd_group)


def list_opts():
    return {
        statsd_group: statsd_opts
    }



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the AlarmProcessor"""

import collections
import json
import time
from unittest import mock

from monasca_common.kafka import legacy_kafka_message

from monasca_notification import notification as m_notification
from monasca_notification.processors import alarm_processor

from tests import base

alarm_tuple = collections.namedtuple('alarm_tuple', ['offset', 'message'])
message_tuple = collections.namedtuple('message_tuple', ['key', 'value'])


class TestAlarmProcessor(base.BaseTestCase):
    def setUp(self):
        super(TestAlarmProcessor, self).setUp()
        self.trap = []

    def _create_raw_alarm(self, partition, offset, message, key=1):
        """Create a raw alarm, with the given message dictionary.
        """
        json_msg = json.dumps({'alarm-transitioned': message})
        msg_tuple = message_tuple(key, json_msg)
        return legacy_kafka_message.LegacyKafkaMessage([partition,
                                                        alarm_tuple(offset,
                                                                    msg_tuple)])

    @mock.patch('pymysql.connect')
    @mock.patch('monasca_notification.processors.alarm_processor.log')
    def _run_alarm_processor(self, alarm, sql_response, mock_log, mock_mysql):
        """Runs a mocked alarm processor reading from queue while running,
        returns (queue_message, log_message)
        """
        # Since the log runs in another thread I can mock it directly, instead
        # change the methods to put to a queue
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        # Setup the sql response
        if sql_response is not None:
            mock_mysql.return_value = mock_mysql
            mock_mysql.cursor.return_value = mock_mysql
            mock_mysql.__iter__.return_value = sql_response

        self.conf_override(group='mysql', ssl=None,
                           host='localhost', port='3306',
                           user='mysql_user', db='dbname',
                           passwd='mysql_passwd')
        self.conf_override(group='statsd', host='localhost',
                           port=8125)
        processor = alarm_processor.AlarmProcessor()

        return processor.to_notification(alarm)

    def test_invalid_alarm(self):
        """Invalid Alarms, should log and error and push to the finished queue."""
        alarm = self._create_raw_alarm(0, 1, {'invalid': 'invalid_alarm'})
        notifications, partition, offset = self._run_alarm_processor(alarm, None)
        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 1)

        invalid_msg = ('Invalid Alarm format skipping partition 0, offset 1\n'
                       'ErrorAlarm data missing field actionsEnabled')

        self.assertIn(invalid_msg, self.trap)

    def test_old_timestamp(self):
        """Should cause the alarm_ttl to fire log a warning and push to finished queue."""
        timestamp = 1375346830042
        alarm_dict = {
            "tenantId": "0",
            "alarmDefinitionId": "0",
            "alarmId": "1",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "stateChangeReason": "I am alarming!",
            "timestamp": timestamp,
            "actionsEnabled": 1,
            "metrics": "cpu_util",
            "severity": "LOW",
            "link": "http://some-place.com",
            "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 2, alarm_dict)
        expected_datetime = time.ctime(timestamp / 1000)

        notifications, partition, offset = self._run_alarm_processor(alarm, None)

        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 2)

        old_msg = ('Received alarm older than the ttl, skipping. '
                   'Alarm from {datetime}'.format(datetime=expected_datetime))

        self.assertIn(old_msg, self.trap)

    def test_no_notifications(self):
        """Test an alarm with no defined notifications
        """
        alarm_dict = {
            "tenantId": "0",
            "alarmDefinitionId": "0",
            "alarmId": "1",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time() * 1000,
            "actionsEnabled": 1,
            "metrics": "cpu_util",
            "severity": "LOW",
            "link": "http://some-place.com",
            "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 3, alarm_dict)

        notifications, partition, offset = self._run_alarm_processor(alarm, None)

        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 3)

    def test_valid_notification(self):
        """Test a valid notification, being put onto the notification_queue
        """
        alarm_dict = {
            "tenantId": "0",
            "alarmDefinitionId": "0",
            "alarmId": "1",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time() * 1000,
            "actionsEnabled": 1,
            "metrics": "cpu_util",
            "severity": "LOW",
            "link": "http://some-place.com",
            "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 4, alarm_dict)

        sql_response = [[1, 'EMAIL', 'test notification', 'me@here.com', 0]]
        notifications, partition, offset = self._run_alarm_processor(alarm, sql_response)

        test_notification = m_notification.Notification(1, 'email', 'test notification',
                                                        'me@here.com', 0, 0, alarm_dict)

        self.assertEqual(notifications, [test_notification])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 4)

    def test_two_valid_notifications(self):
        alarm_dict = {
            "tenantId": "0",
            "alarmDefinitionId": "0",
            "alarmId": "1",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time() * 1000,
            "actionsEnabled": 1,
            "metrics": "cpu_util",
            "severity": "LOW",
            "link": "http://some-place.com",
            "lifecycleState": "OPEN"}

        alarm = self._create_raw_alarm(0, 5, alarm_dict)

        sql_response = [[1, 'EMAIL', 'test notification', 'me@here.com', 0],
                        [2, 'EMAIL', 'test notification2', 'me@here.com', 0]]
        notifications, partition, offset = self._run_alarm_processor(alarm, sql_response)

        test_notification = m_notification.Notification(1, 'email', 'test notification',
                                                        'me@here.com', 0, 0, alarm_dict)
        test_notification2 = m_notification.Notification(2, 'email', 'test notification2',
                                                         'me@here.com', 0, 0, alarm_dict)

        self.assertEqual(notifications, [test_notification, test_notification2])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 5)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from monasca_notification.conf import types

_DEFAULT_URL = '127.0.0.1:2181'
_DEFAULT_NOTIFICATION_PATH = '/notification/alarms'
_DEFAULT_RETRY_PATH = '/notification/retry'
_DEFAULT_PERIODIC_PATH = {
    '60': '/notification/60_seconds'
}

zookeeper_group = cfg.OptGroup('zookeeper',
                               title='Zookeeper Options',
                               help='Options under this group allow to '
                                    'configure settings for zookeeper '
                                    'handling.')

zookeeper_opts = [
    cfg.ListOpt(name='url', item_type=types.HostAddressPortType(),
                default=_DEFAULT_URL, required=True,
                help='List of addresses (with ports) pointing '
                     'at zookeeper cluster.'),
    cfg.StrOpt(name='notification_path', default=_DEFAULT_NOTIFICATION_PATH,
               required=True, advanced=True,
               help='Path in zookeeper tree to track notification offsets.'),
    cfg.StrOpt(name='notification_retry_path', default=_DEFAULT_RETRY_PATH,
               required=True, advanced=True,
               help='Path in zookeeper tree to track notification '
                    'retries offsets.'),
    cfg.DictOpt(name='periodic_path', default=_DEFAULT_PERIODIC_PATH,
                required=True, advanced=True,
                help='Paths in zookeeper tree to track periodic offsets. '
                     'Keys must be integers describing the interval '
                     'of periodic notification. Values are actual '
                     'paths inside zookeeper tree.')
]


def register_opts(conf):
    conf.register_group(zookeeper_group)
    conf.register_opts(zookeeper_opts, group=zookeeper_group)


def list_opts():
    return {
        zookeeper_group: zookeeper_opts
    }



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

_KEY_MAP = {
    'email': 'monasca_notification.plugins.email_notifier:EmailNotifier',
    'pagerduty': 'monasca_notification.plugins.'
                 'pagerduty_notifier:PagerdutyNotifier',
    'webhook': 'monasca_notification.plugins.webhook_notifier:WebhookNotifier',
    'hipchat': 'monasca_notification.plugins.hipchat_notifier:HipChatNotifier',
    'slack': 'monasca_notification.plugins.slack_notifier:SlackNotifier',
    'jira': 'monasca_notification.plugins.jira_notifier:JiraNotifier'
}

notifier_group = cfg.OptGroup('notification_types',
                              title='Notification types',
                              help='Group allows to configure available '
                                   'notifiers inside notification engine.')

notifier_opts = [
    cfg.ListOpt(name='enabled', default=['email', 'pagerduty', 'webhook'],
                item_type=lambda x: _KEY_MAP.get(x, x), bounds=False,
                advanced=True, sample_default=','.join(_KEY_MAP.keys()),
                help='List of enabled notification types. You may specify '
                     'full class name {} '
                     'or shorter label {}.'.format(_KEY_MAP.get('hipchat'),
                                                   'hipchat')
                )
]


def register_opts(conf):
    conf.register_group(notifier_group)
    conf.register_opts(notifier_opts, group=notifier_group)


def list_opts():
    return {
        notifier_group: notifier_opts,
    }



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from monasca_notification import notification as m_notification
from monasca_notification.plugins import hipchat_notifier
import queue
from tests import base


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmDefinitionId": 0,
            "alarmName": "test Alarm",
            "alarmDescription": "test Alarm description",
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "CRITICAL",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": "I am alarming!",
            "timestamp": 1429023453632,
            "metrics": metrics}


class requestsResponse(object):
    def __init__(self, status):
        self.status_code = status


class TestHipchat(base.PluginTestCase):
    def setUp(self):
        super(TestHipchat, self).setUp(hipchat_notifier.register_opts)
        self.conf_default(group='hipchat_notifier', timeout=50)

        self.trap = queue.Queue()

    def tearDown(self):
        super(TestHipchat, self).tearDown()
        self.assertTrue(self.trap.empty())

    def _http_post_200(self, url, data, **kwargs):
        self.trap.put(url)
        self.trap.put(data)
        r = requestsResponse(200)
        return r

    @mock.patch('monasca_notification.plugins.hipchat_notifier.requests')
    def notify(self, http_func, mock_requests):
        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.put
        mock_log.error = self.trap.put
        mock_log.exception = self.trap.put

        mock_requests.post = http_func

        hipchat = hipchat_notifier.HipChatNotifier(mock_log)

        hipchat.config()

        metric = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metric.append(metric_data)

        alarm_dict = alarm(metric)

        notification = m_notification.Notification(0, 'hipchat', 'hipchat notification',
                                                   'http://mock:3333/', 0, 0, alarm_dict)

        self.trap.put(hipchat.send_notification(notification))

    def test_hipchat_success(self):
        """hipchat success
        """
        self.notify(self._http_post_200)

        url = self.trap.get(timeout=1)
        data = self.trap.get(timeout=1)

        self.valid_hipchat_message(url, data)

        return_value = self.trap.get()
        self.assertTrue(return_value)

    def valid_hipchat_message(self, url, data):
        self.assertEqual(url, "http://mock:3333/")

        self.assertEqual(data.get('color'), 'red')
        self.assertEqual(data.get('message_format'), 'text')

        message = json.loads(data.get('message'))
        self.assertEqual(message.get('message'), 'I am alarming!')
        self.assertEqual(message.get('alarm_name'), 'test Alarm')
        self.assertEqual(message.get('alarm_description'), 'test Alarm description')



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

from oslo_config import cfg
from oslo_log import log as logging
import simplejson as json

from monasca_notification.common.repositories import exceptions as exc
from monasca_notification.common.utils import get_db_repo
from monasca_notification.common.utils import get_statsd_client
from monasca_notification import notification
from monasca_notification import notification_exceptions


log = logging.getLogger(__name__)
CONF = cfg.CONF


class AlarmProcessor(object):
    def __init__(self):
        self._alarm_ttl = CONF.alarm_processor.ttl
        self._statsd = get_statsd_client()
        self._db_repo = get_db_repo()

    @staticmethod
    def _parse_alarm(alarm_data):
        """Parse the alarm message making sure it matches the expected format.
        """
        expected_fields = [
            'actionsEnabled',
            'alarmId',
            'alarmDefinitionId',
            'alarmName',
            'newState',
            'oldState',
            'stateChangeReason',
            'severity',
            'link',
            'lifecycleState',
            'tenantId',
            'timestamp'
        ]
        # check if alarm_data is <class 'bytes'>
        # if yes convert it to standard string
        if isinstance(alarm_data, bytes):
            alarm_data = alarm_data.decode("utf-8")
        json_alarm = json.loads(alarm_data)
        alarm = json_alarm['alarm-transitioned']
        for field in expected_fields:
            if field not in alarm:
                raise notification_exceptions.AlarmFormatError(
                    'Alarm data missing field %s' % field)
        if ('tenantId' not in alarm) or ('alarmId' not in alarm):
            raise notification_exceptions.AlarmFormatError

        return alarm

    def _alarm_is_valid(self, alarm):
        """Check if the alarm is enabled and is within the ttl, return True in that case
        """
        if not alarm['actionsEnabled']:
            log.debug('Actions are disabled for this alarm.')
            return False

        alarm_age = time.time() - alarm['timestamp'] / 1000
        if (self._alarm_ttl is not None) and (alarm_age > self._alarm_ttl):
            log.warn('Received alarm older than the ttl, skipping. Alarm from %s' %
                     time.ctime(alarm['timestamp'] / 1000))
            return False

        return True

    def _build_notification(self, alarm):
        db_time = self._statsd.get_timer()

        with db_time.time('config_db_time'):
            alarms_actions = self._db_repo.fetch_notifications(alarm)

        return [notification.Notification(
                alarms_action[0],
                alarms_action[1],
                alarms_action[2],
                alarms_action[3],
                alarms_action[4],
                0,
                alarm) for alarms_action in alarms_actions]

    def to_notification(self, raw_alarm):
        """Check the notification setting for this project then create the appropriate notification
        """
        failed_parse_count = self._statsd.get_counter(name='alarms_failed_parse_count')
        no_notification_count = self._statsd.get_counter(name='alarms_no_notification_count')
        notification_count = self._statsd.get_counter(name='created_count')

        partition = raw_alarm.partition()
        offset = raw_alarm.offset()
        try:
            alarm = self._parse_alarm(raw_alarm.value())
        except Exception as e:  # This is general because of a lack of json exception base class
            failed_parse_count += 1
            log.exception(
                "Invalid Alarm format skipping partition %d, offset %d\nError%s" %
                (partition, offset, e))
            return [], partition, offset

        log.debug("Read alarm from alarms sent_queue. Partition %d, Offset %d, alarm data %s"
                  % (partition, offset, alarm))

        if not self._alarm_is_valid(alarm):
            no_notification_count += 1
            return [], partition, offset

        try:
            notifications = self._build_notification(alarm)
        except exc.DatabaseException:
            log.debug('Database Error.  Attempting reconnect')
            notifications = self._build_notification(alarm)

        if len(notifications) == 0:
            no_notification_count += 1
            log.debug(
                'No notifications found for this alarm, partition %d, offset %d, alarm data %s' %
                (partition, offset, alarm))
            return [], partition, offset
        else:
            log.debug('Found %d notifications: [%s]', len(notifications), notifications)
            notification_count += len(notifications)
            return notifications, partition, offset



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log
import sys

from monasca_notification import conf
from monasca_notification import version

LOG = log.getLogger(__name__)
CONF = conf.CONF
_CONF_LOADED = False


def parse_args(argv):
    """Sets up configuration of monasca-notification."""

    global _CONF_LOADED
    if _CONF_LOADED:
        LOG.debug('Configuration has been already loaded')
        return

    conf.register_opts(CONF)
    log.register_options(CONF)

    default_log_levels = (log.get_default_log_levels())
    log.set_defaults(default_log_levels=default_log_levels)

    CONF(args=argv,
         project='monasca',
         prog=sys.argv[1:],
         version=version.version_string,
         default_config_files=_get_config_files(),
         description='''
         monasca-notification is an engine responsible for
         transforming alarm transitions into proper notifications
         ''')

    conf.register_enabled_plugin_opts(CONF)

    log.setup(CONF,
              product_name='monasca-notification',
              version=version.version_string)

    _CONF_LOADED = True


def _get_config_files():
    """Get the possible configuration files accepted by oslo.config

    This also includes the deprecated ones
    """
    # default files
    conf_files = cfg.find_config_files(project='monasca',
                                       prog='monasca-notification')
    # deprecated config files (only used if standard config files are not there)
    if len(conf_files) == 0:
        old_conf_files = cfg.find_config_files(project='monasca',
                                               prog='notification')
        if len(old_conf_files) > 0:
            LOG.warning('Found deprecated old location "{}" '
                        'of main configuration file'.format(old_conf_files))
            conf_files += old_conf_files
    return conf_files



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

keystone_group = cfg.OptGroup('keystone',
                              title='Keystone Options',
                              help='Options under this group allow to configure '
                                   'valid connection via Keystone'
                                   'authentication.')

keystone_opts = [
    cfg.BoolOpt(name='auth_required', default='False',
                help='This option enable or disable authentication using '
                     'keystone'),
    cfg.StrOpt(name='auth_url', default='http://127.0.0.1/identity/v3',
               help='URL of identity service'),
    cfg.StrOpt(name='username', default='admin',
               help='Username'),
    cfg.StrOpt(name='password', default='password',
               help='Password of identity service'),
    cfg.StrOpt(name='project_name', default='admin',
               help='Name of project'),
    cfg.StrOpt(name='user_domain_name', default='default',
               help='User domain name'),
    cfg.StrOpt(name='project_domain_name', default='default',
               help='Project domain name'),
    cfg.StrOpt(name='auth_type', default='password',
               help='Type of authentication')
]


def register_opts(conf):
    conf.register_group(keystone_group)
    conf.register_opts(keystone_opts, group=keystone_group)


def list_opts():
    return {
        keystone_group: keystone_opts
    }



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
from unittest import mock

import simplejson as json

from monasca_notification import notification as m_notification
from monasca_notification.plugins import webhook_notifier
import queue
from tests import base


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmDefinitionId": 0,
            "alarmName": "test Alarm",
            "alarmDescription": "test Alarm description",
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "LOW",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": "I am alarming!",
            "timestamp": 1429023453632,
            "metrics": metrics}


class requestsResponse(object):
    def __init__(self, status):
        self.status_code = status


class TestWebhook(base.PluginTestCase):

    def setUp(self):
        super(TestWebhook, self).setUp(webhook_notifier.register_opts)
        self.trap = queue.Queue()

    def tearDown(self):
        super(TestWebhook, self).tearDown()
        self.assertTrue(self.trap.empty())

    def _http_post_200(self, url, data, headers, **kwargs):
        self.trap.put(url)
        self.trap.put(data)
        self.trap.put(headers)
        r = requestsResponse(200)
        return r

    def _http_post_404(self, url, data, headers, **kwargs):
        r = requestsResponse(404)
        return r

    def _http_post_exception(self, url, data, headers, **kwargs):
        self.trap.put("timeout %s" % kwargs["timeout"])
        raise requests.exceptions.Timeout

    @mock.patch('monasca_notification.plugins.webhook_notifier.requests')
    def notify(self, http_func, mock_requests):
        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.put
        mock_log.error = self.trap.put
        mock_log.exception = self.trap.put

        mock_requests.post = http_func

        webhook = webhook_notifier.WebhookNotifier(mock_log)

        metric = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metric.append(metric_data)

        alarm_dict = alarm(metric)

        notification = m_notification.Notification(0, 'webhook', 'webhook notification',
                                                   'http://mock:3333/', 0, 0, alarm_dict)

        self.trap.put(webhook.send_notification(notification))

    def test_webhook_good_http_response(self):
        """webhook 200
        """
        self.notify(self._http_post_200)

        url = self.trap.get(timeout=1)
        data = self.trap.get(timeout=1)
        headers = self.trap.get(timeout=1)

        self.assertEqual(url, "http://mock:3333/")
        self.maxDiff = None
        # timestamp is in milliseconds while alarm_timestamp is in seconds
        self.assertEqual(json.loads(data),
                         {"metrics": [{"dimensions": {"hostname": "foo1",
                                                      "service": "bar1"}}],
                          "alarm_id": "0",
                          "state": "ALARM",
                          "alarm_timestamp": 1429023453,
                          "tenant_id": "0",
                          "old_state": "OK",
                          "alarm_description": "test Alarm description",
                          "message": "I am alarming!",
                          "alarm_definition_id": 0,
                          "alarm_name": "test Alarm"})
        self.assertEqual(headers, {'content-type': 'application/json'})

        return_value = self.trap.get()
        self.assertTrue(return_value)

    def test_webhook_bad_http_response(self):
        """webhook bad response
        """

        self.notify(self._http_post_404)

        error = self.trap.get()

        self.assertNotRegex(error, "alarm_id.: .test Alarm")
        self.assertNotRegex(error, "content-type.: .application/json")

        self.assertRegex(error, "HTTP code 404")
        self.assertRegex(error, "post on URL http://mock:3333/")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_webhook_timeout_exception_on_http_response(self):
        """webhook timeout exception
        """

        self.conf_override(group='webhook_notifier', timeout=50)
        self.notify(self._http_post_exception)

        result = self.trap.get()

        self.assertEqual(result, "timeout 50")

        result = self.trap.get()

        self.assertNotRegex(result, "alarm_id.: .test Alarm")
        self.assertNotRegex(result, "content-type.: .application/json")

        self.assertRegex(result, "Error trying to post on URL http://mock:3333/")

        return_value = self.trap.get()
        self.assertFalse(return_value)



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import simplejson as json

from debtcollector import removals
from oslo_config import cfg

from monasca_notification.common import utils
from monasca_notification.plugins import abstract_notifier

CONF = cfg.CONF


class WebhookNotifier(abstract_notifier.AbstractNotifier):

    type = 'webhook'

    def __init__(self, log):
        super(WebhookNotifier, self).__init__()
        self._log = log

    @removals.remove(
        message='Configuration of notifier is available through oslo.cfg',
        version='1.9.0',
        removal_version='3.0.0'
    )
    def config(self, config_dict):
        pass

    @property
    def statsd_name(self):
        return 'sent_webhook_count'

    def send_notification(self, notification):
        """Send the notification via webhook
            Posts on the given url
        """

        body = {'alarm_id': notification.alarm_id,
                'alarm_definition_id': notification.raw_alarm['alarmDefinitionId'],
                'alarm_name': notification.alarm_name,
                'alarm_description': notification.raw_alarm['alarmDescription'],
                'alarm_timestamp': notification.alarm_timestamp,
                'state': notification.state,
                'old_state': notification.raw_alarm['oldState'],
                'message': notification.message,
                'tenant_id': notification.tenant_id,
                'metrics': notification.metrics}

        headers = {'content-type': 'application/json'}

        # Checks if keystone authentication is enabled and adds authentication
        # token to the request headers
        if CONF.keystone.auth_required:
            auth_token = utils.get_auth_token()
            headers = {'content-type': 'application/json',
                       'X-Auth-Token': auth_token}

        url = notification.address

        try:
            # Posting on the given URL
            result = requests.post(url=url,
                                   data=json.dumps(body),
                                   headers=headers,
                                   timeout=CONF.webhook_notifier.timeout)

            if result.status_code in range(200, 300):
                self._log.info("Notification successfully posted.")
                return True
            else:
                self._log.error("Received an HTTP code {} when trying to post on URL {}."
                                .format(result.status_code, url))
                return False
        except Exception:
            self._log.exception("Error trying to post on URL {}".format(url))
            return False


webhook_notifier_group = cfg.OptGroup(name='%s_notifier' % WebhookNotifier.type)
webhook_notifier_opts = [
    cfg.IntOpt(name='timeout', default=5, min=1)
]


def register_opts(conf):
    conf.register_group(webhook_notifier_group)
    conf.register_opts(webhook_notifier_opts, group=webhook_notifier_group)


def list_opts():
    return {
        webhook_notifier_group: webhook_notifier_opts
    }



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import fixtures

from oslo_config import cfg
from oslo_config import fixture as oo_cfg
from oslotest import base as oslotest_base

from monasca_notification import conf
from monasca_notification import config


class DisableStatsdFixture(fixtures.Fixture):

    def setUp(self):
        super(DisableStatsdFixture, self).setUp()
        statsd_patch = mock.patch('monascastatsd.Connection')
        statsd_patch.start()
        self.addCleanup(statsd_patch.stop)


class ConfigFixture(oo_cfg.Config):
    """Mocks configuration"""

    def __init__(self):
        super(ConfigFixture, self).__init__(config.CONF)

    def setUp(self):
        super(ConfigFixture, self).setUp()

        self.addCleanup(self._clean_config_loaded_flag)

        conf.register_opts()
        config.parse_args(argv=[])

    @staticmethod
    def _clean_config_loaded_flag():
        config._CONF_LOADED = False


class BaseTestCase(oslotest_base.BaseTestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(ConfigFixture())
        self.useFixture(DisableStatsdFixture())

    @staticmethod
    def conf_override(**kw):
        """Override flag variables for a test."""
        group = kw.pop('group', None)
        for k, v in kw.items():
            cfg.CONF.set_override(k, v, group)

    @staticmethod
    def conf_default(**kw):
        """Override flag variables for a test."""
        group = kw.pop('group', None)
        for k, v in kw.items():
            cfg.CONF.set_default(k, v, group)


class PluginTestCase(BaseTestCase):
    register_opts = None

    def setUp(self, register_opts=None):
        super(PluginTestCase, self).setUp()
        if register_opts:
            register_opts(conf.CONF)



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import simplejson as json
import urllib

from debtcollector import removals
from oslo_config import cfg

from monasca_notification.plugins import abstract_notifier

CONF = cfg.CONF

"""
notification.address = https://hipchat.hpcloud.net/v2/room/<room_id>/notification?auth_token=432432

How to get access token?
    1) Login to Hipchat with the user account which is used for notification
    2) Go to this page. https://hipchat.hpcloud.net/account/api (Replace your hipchat server name)
    3) You can see option to "Create token". Use the capability "SendNotification"

How to get the Room ID?
    1) Login to Hipchat with the user account which is used for notification
    2) Go to this page. https://hipchat.hpcloud.net/account/api (Replace your hipchat server name)
    3) Click on the Rooms tab
    4) Click on any Room of your choice.
    5) Room ID is the API ID field

"""

SEVERITY_COLORS = {"low": 'green',
                   'medium': 'gray',
                   'high': 'yellow',
                   'critical': 'red'}


class HipChatNotifier(abstract_notifier.AbstractNotifier):

    type = 'hipchat'

    def __init__(self, log):
        super(HipChatNotifier, self).__init__()
        self._log = log

    @removals.remove(
        message='Configuration of notifier is available through oslo.cfg',
        version='1.9.0',
        removal_version='3.0.0'
    )
    def config(self, config_dict=None):
        pass

    @property
    def statsd_name(self):
        return 'sent_hipchat_count'

    def _build_hipchat_message(self, notification):
        """Builds hipchat message body
        """
        body = {'alarm_id': notification.alarm_id,
                'alarm_definition_id': notification.raw_alarm['alarmDefinitionId'],
                'alarm_name': notification.alarm_name,
                'alarm_description': notification.raw_alarm['alarmDescription'],
                'alarm_timestamp': notification.alarm_timestamp,
                'state': notification.state,
                'old_state': notification.raw_alarm['oldState'],
                'message': notification.message,
                'tenant_id': notification.tenant_id,
                'metrics': notification.metrics}

        hipchat_request = {}
        hipchat_request['color'] = self._get_color(notification.severity.lower())
        hipchat_request['message_format'] = 'text'
        hipchat_request['message'] = json.dumps(body, indent=3)

        return hipchat_request

    def _get_color(self, severity):
        return SEVERITY_COLORS.get(severity, 'purple')

    def send_notification(self, notification):
        """Send the notification via hipchat
            Posts on the given url
        """

        hipchat_message = self._build_hipchat_message(notification)
        parsed_url = urllib.parse.urlsplit(notification.address)

        query_params = urllib.parse.parse_qs(parsed_url.query)
        # URL without query params
        url = urllib.parse.urljoin(
            notification.address,
            urllib.parse.urlparse(
                notification.address).path)

        # Default option is to do cert verification
        verify = not CONF.hipchat_notifier.insecure
        ca_certs = CONF.hipchat_notifier.ca_certs
        proxy = CONF.hipchat_notifier.proxy

        # If ca_certs is specified, do cert validation and ignore insecure flag
        if ca_certs is not None:
            verify = ca_certs

        proxyDict = None
        if proxy is not None:
            proxyDict = {'https': proxy}

        try:
            # Posting on the given URL
            result = requests.post(url=url,
                                   data=hipchat_message,
                                   verify=verify,
                                   params=query_params,
                                   proxies=proxyDict,
                                   timeout=CONF.hipchat_notifier.timeout)

            if result.status_code in range(200, 300):
                self._log.info("Notification successfully posted.")
                return True
            else:
                msg = ("Received an HTTP code {} when trying to send to hipchat on URL {}"
                       " with response {}.")
                self._log.error(msg.format(result.status_code, url, result.text))
                return False
        except Exception:
            self._log.exception("Error trying to send to hipchat on URL {}".format(url))
            return False


hipchat_notifier_group = cfg.OptGroup(name='%s_notifier' % HipChatNotifier.type)
hipchat_notifier_opts = [
    cfg.IntOpt(name='timeout', default=5, min=1),
    cfg.BoolOpt(name='insecure', default=True),
    cfg.StrOpt(name='ca_certs', default=None),
    cfg.StrOpt(name='proxy', default=None)
]


def register_opts(conf):
    conf.register_group(hipchat_notifier_group)
    conf.register_opts(hipchat_notifier_opts, group=hipchat_notifier_group)


def list_opts():
    return {
        hipchat_notifier_group: hipchat_notifier_opts
    }



# coding=utf-8
# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import email.header
import smtplib
import socket
import time
from unittest import mock
from urllib import parse
from urllib.parse import urlparse

from monasca_notification.notification import Notification
from monasca_notification.plugins import email_notifier
from tests import base


UNICODE_CHAR = chr(2344)
UNICODE_CHAR_ENCODED = UNICODE_CHAR.encode("utf-8")


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmName": u"test Alarm " + UNICODE_CHAR,
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "LOW",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": u"I am alarming!" + UNICODE_CHAR,
            "timestamp": time.time(),
            "metrics": metrics}


def _parse_email(email_msg):
    raw_mail = {"raw": email_msg}
    email_lines = email_msg.splitlines()

    from_addr, subject, to_addr = _decode_headers(email_lines)

    raw_mail['subject'] = subject[0].decode(subject[1])
    raw_mail['from'] = from_addr[0]
    raw_mail['to'] = to_addr[0]
    raw_mail['body'] = (base64.b64decode('\n'.join(email_lines[8:]))
                        .decode('utf-8'))

    return raw_mail


def _decode_headers(email_lines):
    # message is encoded, so we need to carefully go through all the lines
    # to pick ranges for subject, from and to
    keys = ['Subject', 'From', 'To']
    subject, from_addr, to_addr = None, None, None
    for key_idx, key in enumerate(keys):
        accummulated = []
        for idx in range(3, len(email_lines) - 1):
            line = email_lines[idx]
            if not line:
                break
            if key in line:
                accummulated.append(line)
                try:
                    if keys[key_idx + 1] not in email_lines[idx + 1]:
                        accummulated.append(email_lines[idx + 1])
                    else:
                        break
                except IndexError:
                    pass
        if key == 'Subject':
            subject = email.header.decode_header(''.join(accummulated))[1]
        if key == 'From':
            from_addr = email.header.decode_header(''.join(accummulated))[0]
        if key == 'To':
            to_addr = email.header.decode_header(''.join(accummulated))[0]
    return from_addr, subject, to_addr


class smtpStub(object):
    def __init__(self, trap):
        self.trap = trap

    def sendmail(self, from_addr, to_addr, msg):
        self.trap.append("{} {} {}".format(from_addr, to_addr, msg))


class smtpStubException(object):
    def __init__(self, queue):
        self.queue = queue

    def sendmail(self, from_addr, to_addr, msg):
        raise smtplib.SMTPServerDisconnected


class TestEmail(base.PluginTestCase):
    def setUp(self):
        super(TestEmail, self).setUp(email_notifier.register_opts)

        self.trap = []
        self.conf_override(group='email_notifier', server='my.smtp.server',
                           port=25, user=None,
                           password=None, timeout=60,
                           from_addr='hpcs.mon@hp.com',
                           grafana_url='http://127.0.0.1:3000')

    def _smtpStub(self, *arg, **kwargs):
        return smtpStub(self.trap)

    def _smtbStubException(self, *arg, **kwargs):
        return smtpStubException(self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def notify(self, smtp_stub, metric, mock_smtp):
        mock_smtp.SMTP = smtp_stub

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append

        email = email_notifier.EmailNotifier(mock_log)

        alarm_dict = alarm(metric)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        self.trap.append(email.send_notification(notification))

    def test_email_notification_single_host(self):
        """Email with single host
        """

        metrics = []
        metric_data = {'name': 'cpu.percent',
                       'dimensions': {'hostname': u'foo1' + UNICODE_CHAR,
                                      u'service' + UNICODE_CHAR: 'bar1'}}
        metrics.append(metric_data)

        self.notify(self._smtpStub, metrics)

        email = _parse_email(self.trap.pop(0))

        self.assertRegex(email['from'], 'hpcs.mon@hp.com')
        self.assertRegex(email['to'], 'me@here.com')
        self.assertRegex(email['raw'], 'Content-Type: text/plain')
        self.assertRegex(email['raw'], 'Content-Transfer-Encoding: base64')
        self.assertRegex(email['subject'],
                         'ALARM LOW "test Alarm .*" for Host: foo1.*')
        self.assertRegex(email['body'], 'Alarm .test Alarm.')
        self.assertRegex(email['body'], 'On host .foo1.')
        self.assertRegex(email['body'], UNICODE_CHAR)
        self.assertRegex(email['body'], 'Link: some-link')
        self.assertRegex(email['body'], 'Lifecycle state: OPEN')

        return_value = self.trap.pop(0)
        self.assertTrue(return_value)

    def test_email_notification_target_host(self):
        """Email with single host
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': u'foo1' + UNICODE_CHAR,
                                                             u'service' + UNICODE_CHAR: 'bar1',
                                                             u'target_host': u'some_where'}}
        metrics.append(metric_data)

        self.notify(self._smtpStub, metrics)

        email = _parse_email(self.trap.pop(0))

        self.assertRegex(email['from'], 'hpcs.mon@hp.com')
        self.assertRegex(email['to'], 'me@here.com')
        self.assertRegex(email['raw'], 'Content-Type: text/plain')
        self.assertRegex(email['raw'], 'Content-Transfer-Encoding: base64')
        self.assertRegex(email['subject'],
                         'ALARM LOW .test Alarm.* Target: some_where')
        self.assertRegex(email['body'], "Alarm .test Alarm.")
        self.assertRegex(email['body'], "On host .foo1.")
        self.assertRegex(email['body'], UNICODE_CHAR)

        return_value = self.trap.pop(0)
        self.assertTrue(return_value)

    def worktest_email_notification_multiple_hosts(self):
        """Email with multiple hosts
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        self.notify(self._smtpStub, metrics)

        email = _parse_email(self.trap.pop(0))

        self.assertRegex(email['from'], "From: hpcs.mon@hp.com")
        self.assertRegex(email['to'], "To: me@here.com")
        self.assertRegex(email['raw'], "Content-Type: text/plain")
        self.assertRegex(email['subject'], "Subject: ALARM LOW .test Alarm.")
        self.assertRegex(email['body'], "Alarm .test Alarm.")
        self.assertRegex(email['body'], "foo1")
        self.assertRegex(email['body'], "foo2")
        self.assertRegex(email['body'], "bar1")
        self.assertRegex(email['body'], "bar2")

        return_value = self.trap.pop(0)
        self.assertTrue(return_value)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_smtp_sendmail_failed_connection_twice(self, mock_smtp):
        """Email that fails on smtp_connect twice
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        # mock_smtp.SMTP.return_value = mock_smtp
        mock_smtp.SMTP.side_effect = [mock_smtp,
                                      smtplib.SMTPServerDisconnected,
                                      socket.error]

        mock_smtp.sendmail.side_effect = [smtplib.SMTPServerDisconnected,
                                          smtplib.SMTPServerDisconnected]

        # There has to be a better way to preserve exception definitions when
        # we're mocking access to a library
        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        email.config()

        alarm_dict = alarm(metrics)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        self.trap.append(email.send_notification(notification))

        self.assertIn("SMTP server disconnected. Will reconnect and retry message.", self.trap)
        self.assertIn("Unable to connect to email server.", self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_smtp_sendmail_smtp_None(self, mock_smtp):
        """Email that fails on smtp_connect twice
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        mock_smtp.SMTP.return_value = None
        mock_smtp.SMTP.side_effect = [socket.error,
                                      socket.error,
                                      socket.error]

        mock_smtp.sendmail.side_effect = [smtplib.SMTPServerDisconnected,
                                          smtplib.SMTPServerDisconnected]

        # There has to be a better way to preserve exception definitions when
        # we're mocking access to a library
        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        email.config()

        del self.trap[:]

        alarm_dict = alarm(metrics)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        email_result = email.send_notification(notification)

        self.assertFalse(email_result)
        self.assertIn("Connecting to Email Server my.smtp.server",
                      self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_smtp_sendmail_failed_connection_once_then_email(self, mock_smtp):
        """Email that fails on smtp_connect once then email
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        mock_smtp.SMTP.return_value = mock_smtp

        mock_smtp.sendmail.side_effect = [smtplib.SMTPServerDisconnected,
                                          smtplib.SMTPException]

        # There has to be a better way to preserve exception definitions when
        # we're mocking access to a library
        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        email.config()

        alarm_dict = alarm(metrics)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        self.trap.append(email.send_notification(notification))

        self.assertIn("SMTP server disconnected. Will reconnect and retry message.", self.trap)
        self.assertIn("Error sending Email Notification", self.trap)
        self.assertNotIn("Unable to connect to email server.", self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_smtp_sendmail_failed_connection_once(self, mock_smtp):
        """Email that fails on smtp_connect once
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        mock_smtp.SMTP.return_value = mock_smtp
        mock_smtp.sendmail.side_effect = [smtplib.SMTPServerDisconnected, None]

        # There has to be a better way to preserve exception definitions when
        # we're mocking access to a library
        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        email.config()

        alarm_dict = alarm(metrics)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        self.trap.append(email.send_notification(notification))

        self.assertIn("SMTP server disconnected. Will reconnect and retry message.", self.trap)
        self.assertIn("Sent email to %s, notification %s"
                      % (notification.address, notification.to_json()), self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_smtp_sendmail_failed_exception(self, mock_smtp):
        """Email that fails on exception
        """

        metrics = []
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric_data)
        metric_data = {'name': 'cpu.percent', 'dimensions': {'hostname': 'foo2', 'service': 'bar2'}}
        metrics.append(metric_data)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        mock_smtp.SMTP.return_value = mock_smtp
        mock_smtp.sendmail.side_effect = smtplib.SMTPException

        # There has to be a better way to preserve exception definitions when
        # we're mocking access to a library
        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        email.config()

        alarm_dict = alarm(metrics)

        notification = Notification(0, 'email', 'email notification',
                                    'me@here.com', 0, 0, alarm_dict)

        self.trap.append(email.send_notification(notification))

        self.assertNotIn("SMTP server disconnected. Will reconnect and retry message.", self.trap)
        self.assertIn("Error sending Email Notification", self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    def test_get_link_url(self, mock_smtp):
        # Given one metric with name and dimensions
        metrics = []
        metric = {'name': 'cpu.percent',
                  'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metrics.append(metric)

        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.debug = self.trap.append
        mock_log.info = self.trap.append
        mock_log.exception = self.trap.append

        mock_smtp.SMTP.return_value = mock_smtp
        mock_smtp.sendmail.side_effect = smtplib.SMTPException

        mock_smtp.SMTPServerDisconnected = smtplib.SMTPServerDisconnected
        mock_smtp.SMTPException = smtplib.SMTPException

        email = email_notifier.EmailNotifier(mock_log)

        # Create alarm timestamp and timestamp for 'from' and 'to' dates in milliseconds.
        alarm_date = datetime.datetime(2017, 6, 7, 18, 0)
        alarm_ms, expected_from_ms, expected_to_ms = self.create_time_data(alarm_date)

        # When retrieving the link to Grafana for the first metric and given timestamp
        result_url = email._get_link_url(metrics[0], alarm_ms)
        self.assertIsNotNone(result_url)

        # Then the following link to Grafana (including the metric info and timestamp) is expected.
        expected_url = "http://127.0.0.1:3000/dashboard/script/drilldown.js" \
                       "?metric=cpu.percent&hostname=foo1&service=bar1" \
                       "&from=%s&to=%s" % (expected_from_ms, expected_to_ms)
        self._assert_equal_urls(expected_url, result_url)

    def create_time_data(self, alarm_date):
        epoch = datetime.datetime.utcfromtimestamp(0)
        alarm_ms = int(round((alarm_date - epoch).total_seconds() * 1000))

        # From and to dates are 10 minutes before and after the alarm occurred.
        from_date = alarm_date - datetime.timedelta(minutes=10)
        to_date = alarm_date + datetime.timedelta(minutes=10)

        expected_from_ms = int(round((from_date - epoch).total_seconds() * 1000))
        expected_to_ms = int(round((to_date - epoch).total_seconds() * 1000))

        return alarm_ms, expected_from_ms, expected_to_ms

    def _assert_equal_urls(self, expected_url, result_url):
        expected_parsed = urlparse(expected_url)
        result_parsed = urlparse(result_url)

        self.assertEqual(expected_parsed.netloc, result_parsed.netloc)
        self.assertEqual(expected_parsed.path, result_parsed.path)

        expected_parsed_query = parse.parse_qs(expected_parsed.query)
        result_parsed_query = parse.parse_qs(result_parsed.query)

        self.assertEqual(len(expected_parsed_query), len(result_parsed_query))

        for key in result_parsed_query.keys():
            self.assertEqual(expected_parsed_query[key], result_parsed_query[key])



# Copyright 2015-2017 FUJITSU LIMITED
# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

from oslo_log import log as logging
import psycopg2

from monasca_notification.common.repositories.base import base_repo
from monasca_notification.common.repositories import exceptions as exc

log = logging.getLogger(__name__)


class PostgresqlRepo(base_repo.BaseRepo):
    def __init__(self, config):
        super(PostgresqlRepo, self).__init__(config)
        self._pgsql_params = config['postgresql']
        self._pgsql = None

    def _connect_to_pgsql(self):
        self._pgsql = None
        try:
            self._pgsql = psycopg2.connect(**self._pgsql_params)
            self._pgsql.autocommit = True
        except psycopg2.Error as e:
            log.exception('Pgsql connect failed %s', e)
            raise

    def fetch_notifications(self, alarm):
        try:
            if self._pgsql is None:
                self._connect_to_pgsql()
            cur = self._pgsql.cursor()
            cur.execute(
                self._find_alarm_action_sql,
                (alarm['alarmDefinitionId'],
                 alarm['newState']))
            for row in cur:
                yield (row[0], row[1].lower(), row[2], row[3], row[4])
        except psycopg2.Error as e:
            log.exception("Couldn't fetch alarms actions %s", e)
            raise exc.DatabaseException(e)

    def get_alarm_current_state(self, alarm_id):
        try:
            if self._pgsql is None:
                self._connect_to_pgsql()
            cur = self._pgsql.cursor()
            cur.execute(self._find_alarm_state_sql, alarm_id)
            row = cur.fetchone()
            state = row[0] if row is not None else None
            return state
        except psycopg2.Error as e:
            log.exception("Couldn't fetch current alarm state %s", e)
            raise exc.DatabaseException(e)

    def fetch_notification_method_types(self):
        try:
            if self._pgsql is None:
                self._connect_to_pgsql()
            cur = self._pgsql.cursor()
            cur.execute(self._find_all_notification_types_sql)
            for row in cur:
                yield (row[0])
        except psycopg2.Error as e:
            log.exception("Couldn't fetch notification types %s", e)
            raise exc.DatabaseException(e)

    def insert_notification_method_types(self, notification_types):
        try:
            if self._pgsql is None:
                self._connect_to_pgsql()
            cur = self._pgsql.cursor()
            cur.executemany(self._insert_notification_types_sql, notification_types)
        except psycopg2.Error as e:
            log.exception("Couldn't insert notification types %s", e)
            raise exc.DatabaseException(e)

    def get_notification(self, notification_id):
        try:
            if self._pgsql is None:
                self._connect_to_pgsql()
            cur = self._pgsql.cursor()
            cur.execute(self._get_notification_sql, notification_id)
            row = cur.fetchone()
            if row is None:
                return None
            else:
                return [row[0], row[1].lower(), row[2], row[3]]
        except psycopg2.Error as e:
            log.exception("Couldn't fetch the notification method %s", e)
            raise exc.DatabaseException(e)



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from debtcollector import removals
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

log = logging.getLogger(__name__)
CONF = cfg.CONF

possible_notifiers = None
configured_notifiers = None
statsd_counter = None

statsd = None
statsd_timer = None


def init(statsd_obj):
    global statsd, statsd_timer, \
        possible_notifiers, configured_notifiers,\
        statsd_counter

    statsd = statsd_obj
    statsd_timer = statsd.get_timer()

    statsd_counter = {}
    configured_notifiers = {}
    possible_notifiers = []


def load_plugins():
    global possible_notifiers
    for plugin_class in CONF.notification_types.enabled:
        try:
            plugin_class = plugin_class.replace(':', '.')
            clz = importutils.import_class(plugin_class)
            possible_notifiers.append(clz(logging.getLogger(plugin_class)))
        except Exception:
            log.exception("unable to load the class %s , ignoring it" %
                          plugin_class)


def enabled_notifications():
    global configured_notifiers

    results = []
    for key in configured_notifiers:
        results.append(key.upper())
    return results


@removals.remove(
    message='Loading the plugin configuration has been moved to oslo, '
            'This method will be fully deleted in future releases',
    version='1.9.0',
    removal_version='3.0.0'
)
def config():
    global possible_notifiers, configured_notifiers, statsd_counter

    for notifier in possible_notifiers:
        ntype = notifier.type.lower()
        configured_notifiers[ntype] = notifier
        statsd_counter[ntype] = statsd.get_counter(notifier.statsd_name)
        log.info("{} notification ready".format(ntype))


def send_notifications(notifications):
    sent = []
    failed = []
    invalid = []

    for notification in notifications:
        ntype = notification.type
        if ntype not in configured_notifiers:
            log.warn("attempting to send unconfigured notification: {}".format(ntype))
            invalid.append(notification)
            continue

        notification.notification_timestamp = time.time()

        with statsd_timer.time(ntype + '_time'):
            result = send_single_notification(notification)

        if result:
            sent.append(notification)
            statsd_counter[ntype].increment(1)
        else:
            failed.append(notification)

    return sent, failed, invalid


def send_single_notification(notification):
    global configured_notifiers

    ntype = notification.type
    try:
        return configured_notifiers[ntype].send_notification(notification)
    except Exception:
        log.exception("send_notification exception for {}".format(ntype))
        return False



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

queues_group = cfg.OptGroup('queues',
                            title='Queues Options',
                            help=('Options under this group allow to '
                                  'configure valid connection sizes of '
                                  'all queues.'))

queues_opts = [
    cfg.IntOpt(name='alarms_size', min=1, default=256,
               help='Size of the alarms queue.'),
    cfg.IntOpt(name='finished_size', min=1, default=256,
               help='Size of the finished alarms queue.'),
    cfg.IntOpt(name='notifications_size', min=1, default=256,
               help='Size of notifications queue.'),
    cfg.IntOpt(name='sent_notifications_size', min=1, default=50,
               help='Size of sent notifications queue. '
                    'Limiting this size reduces potential or '
                    're-sent notifications after a failure.')
]


def register_opts(conf):
    conf.register_group(queues_group)
    conf.register_opts(queues_opts, group=queues_group)


def list_opts():
    return {
        queues_group: queues_opts
    }



# Copyright 2015 FUJITSU LIMITED
# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.


class BaseRepo(object):
    def __init__(self, config):
        self._find_alarm_action_sql = \
            """SELECT id, type, name, address, period
               FROM alarm_action as aa
               JOIN notification_method as nm ON aa.action_id = nm.id
               WHERE aa.alarm_definition_id = %s and aa.alarm_state = %s"""
        self._find_alarm_state_sql = \
            """SELECT state
               FROM alarm
               WHERE alarm.id = %s"""
        self._insert_notification_types_sql = \
            """INSERT INTO notification_method_type (name) VALUES ( %s)"""
        self._find_all_notification_types_sql = """SELECT name from notification_method_type """
        self._get_notification_sql = """SELECT name, type, address, period
                                        FROM notification_method
                                        WHERE id = %s"""



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_config import types
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import netutils

LOG = log.getLogger(__name__)


class Plugin(types.String):
    def __init__(self, ignore_missing=False, choices=None, plugin_map=None):

        if not plugin_map:
            # since simport is used, we cannot tell where plugin is located
            # thus plugin map wouldn't contain it
            plugin_map = {}

        super(Plugin, self).__init__(choices, quotes=False, ignore_case=True,
                                     type_name='plugin value')
        self._plugin_map = plugin_map
        self._ingore_mission = ignore_missing

    def __call__(self, value):
        value = super(Plugin, self).__call__(value)
        value = self._get_actual_target(value)
        cls = None

        try:
            value = value.replace(':', '.')
            cls = importutils.import_class(value)
        except ImportError:
            if not self._ingore_mission:
                raise ValueError('%s cannot be imported' % value)
            else:
                LOG.exception('%s cannot be imported', value)

        return cls

    def _get_actual_target(self, value):

        # NOTE(trebskit) missing values will be handled
        # by choices from StringType

        if value in self._plugin_map.keys():
            return self._plugin_map[value]

        return value


class PluginOpt(cfg.Opt):
    def __init__(self, name, choices=None, plugin_map=None, **kwargs):
        plugin = Plugin(choices=choices, plugin_map=plugin_map)
        super(PluginOpt, self).__init__(name,
                                        type=plugin,
                                        **kwargs)


class HostAddressPortType(types.HostAddress):
    """HostAddress with additional port"""

    def __init__(self, version=None):
        type_name = 'host address port value'
        super(HostAddressPortType, self).__init__(version,
                                                  type_name=type_name)

    def __call__(self, value):
        addr, port = netutils.parse_host_port(value)

        addr = self._validate_addr(addr)
        port = self._validate_port(port)
        LOG.debug('addr: %s port: %s' % (addr, port))

        if addr and port:
            return '%s:%d' % (addr, port)
        raise ValueError('%s is not valid host address with optional port')

    @staticmethod
    def _validate_port(port=80):
        port = types.Port()(port)
        return port

    def _validate_addr(self, addr):
        try:
            addr = self.ip_address(addr)
        except ValueError:
            try:
                addr = self.hostname(addr)
            except ValueError:
                raise ValueError("%s is not a valid host address", addr)
        return addr



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from monasca_notification import notification as m_notification
from monasca_notification.plugins import slack_notifier
import queue
from tests import base


def alarm(metrics):
    return {'tenantId': '0',
            'alarmId': '0',
            'alarmDefinitionId': 0,
            'alarmName': 'test Alarm',
            'alarmDescription': 'test Alarm description',
            'oldState': 'OK',
            'newState': 'ALARM',
            'severity': 'CRITICAL',
            'link': 'some-link',
            'lifecycleState': 'OPEN',
            'stateChangeReason': 'I am alarming!',
            'timestamp': 1429023453632,
            'metrics': metrics}


def slack_text():
    return {'old_state': 'OK',
            'alarm_description': 'test Alarm description',
            'message': 'I am alarming!',
            'alarm_definition_id': 0,
            'alarm_name': 'test Alarm',
            'tenant_id': '0',
            'metrics': [
                {'dimensions': {
                    'hostname': 'foo1',
                    'service': 'bar1'}}
            ],
            'alarm_id': '0',
            'state': 'ALARM',
            'alarm_timestamp': 1429023453}


class RequestsResponse(object):
    def __init__(self, status, text, headers):
        self.status_code = status
        self.text = text
        self.headers = headers

    def json(self):
        return json.loads(self.text)


class TestSlack(base.PluginTestCase):

    def setUp(self):
        super(TestSlack, self).setUp(
            slack_notifier.register_opts
        )
        self.conf_default(group='slack_notifier', timeout=50,
                          ca_certs='/etc/ssl/certs/ca-bundle.crt',
                          proxy='http://yourid:password@proxyserver:8080',
                          insecure=False)

        self._trap = queue.Queue()

        mock_log = mock.Mock()
        mock_log.info = self._trap.put
        mock_log.warn = self._trap.put
        mock_log.error = self._trap.put
        mock_log.exception = self._trap.put

        self._slk = slack_notifier.SlackNotifier(mock_log)
        slack_notifier.SlackNotifier._raw_data_url_caches = []

    @mock.patch('monasca_notification.plugins.slack_notifier.requests')
    def _notify(self, response_list, mock_requests):
        mock_requests.post = mock.Mock(side_effect=response_list)

        metric = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metric.append(metric_data)

        alarm_dict = alarm(metric)

        notification = m_notification.Notification(0, 'slack', 'slack notification',
                                                   'http://test.slack:3333', 0, 0,
                                                   alarm_dict)

        return mock_requests.post, self._slk.send_notification(notification)

    def _validate_post_args(self, post_args, data_format):
        self.assertEqual(slack_text(),
                         json.loads(post_args.get(data_format).get('text')))
        self._validate_post_args_base(post_args)

    def _validate_templated_post_args(self, post_args, data_format,
                                      expected_slack_text):
        self.assertEqual(expected_slack_text,
                         post_args.get(data_format).get('text'))
        self._validate_post_args_base(post_args)

    def _validate_post_args_base(self, post_args):
        self.assertEqual({'https': 'http://yourid:password@proxyserver:8080'},
                         post_args.get('proxies'))
        self.assertEqual(50, post_args.get('timeout'))
        self.assertEqual('http://test.slack:3333', post_args.get('url'))
        self.assertEqual('/etc/ssl/certs/ca-bundle.crt',
                         post_args.get('verify'))

    def test_slack_webhook_success(self):
        """slack success
        """
        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]
        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        self._validate_post_args(mock_method.call_args_list[0][1], 'json')
        self.assertEqual([], slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_templated_slack_webhook_success(self):
        """A templated message is successfully sent as json
        """
        self.conf_override(
            message_template='tests/resources/test_slackformat.template',
            group='slack_notifier'
        )
        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]
        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        expected_text = 'test Alarm has triggered on host foo1, service bar1.'
        self._validate_templated_post_args(mock_method.call_args_list[0][1],
                                           'json',
                                           expected_text)
        self.assertEqual([], slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_slack_webhook_fail(self):
        """data is sent twice as json and raw data, and slack returns failure for
           both requests
        """
        response_list = [RequestsResponse(200, 'failure',
                                          {'Content-Type': 'application/text'}),
                         RequestsResponse(200, '{"ok":false,"error":"failure"}',
                                          {'Content-Type': 'application/json'})]
        mock_method, result = self._notify(response_list)
        self.assertFalse(result)
        self._validate_post_args(mock_method.call_args_list[0][1], 'json')
        self._validate_post_args(mock_method.call_args_list[1][1], 'data')
        self.assertEqual([], slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_slack_post_message_success_no_cache(self):
        """data is sent as json at first and get error, second it's sent as raw data
        """
        response_list = [RequestsResponse(200, '{"ok":false,"error":"failure"}',
                                          {'Content-Type': 'application/json'}),
                         RequestsResponse(200, '{"ok":true}',
                                          {'Content-Type': 'application/json'})]
        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        self._validate_post_args(mock_method.call_args_list[0][1], 'json')
        self._validate_post_args(mock_method.call_args_list[1][1], 'data')
        self.assertEqual(['http://test.slack:3333'],
                         slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_slack_post_message_success_cached(self):
        """url in cache and data is sent as raw data at first time
        """
        with mock.patch.object(slack_notifier.SlackNotifier,
                               '_raw_data_url_caches',
                               ['http://test.slack:3333']):
            response_list = [RequestsResponse(200, '{"ok":true}',
                                              {'Content-Type': 'application/json'})]
            mock_method, result = self._notify(response_list)
            self.assertTrue(result)
            mock_method.assert_called_once()
            self._validate_post_args(mock_method.call_args_list[0][1], 'data')
            self.assertEqual(['http://test.slack:3333'],
                             slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_slack_post_message_failed_cached(self):
        """url in cache and slack returns failure
        """
        with mock.patch.object(slack_notifier.SlackNotifier,
                               '_raw_data_url_caches',
                               ['http://test.slack:3333']):
            response_list = [RequestsResponse(200, '{"ok":false,"error":"failure"}',
                                              {'Content-Type': 'application/json'})]
            mock_method, result = self._notify(response_list)
            self.assertFalse(result)
            mock_method.assert_called_once()
            self._validate_post_args(mock_method.call_args_list[0][1], 'data')
            self.assertEqual(['http://test.slack:3333'],
                             slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_slack_webhook_success_only_timeout(self):
        """slack success with only timeout config
        """
        self.conf_override(group='slack_notifier', timeout=50,
                           insecure=True, ca_certs=None,
                           proxy=None)

        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]
        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        self.assertEqual(slack_notifier.SlackNotifier._raw_data_url_caches, [])

        post_args = mock_method.call_args_list[0][1]
        self.assertEqual(slack_text(),
                         json.loads(post_args.get('json').get('text')))
        self.assertEqual(None, post_args.get('proxies'))
        self.assertEqual(50, post_args.get('timeout'))
        self.assertEqual('http://test.slack:3333', post_args.get('url'))
        self.assertFalse(post_args.get('verify'))

    def test_slack_reponse_400(self):
        """slack returns 400 error
        """
        response_list = [RequestsResponse(400, '{"ok":false,"error":"failure"}',
                                          {'Content-Type': 'application/json'}),
                         RequestsResponse(400, '{"ok":false,"error":"failure"}',
                                          {'Content-Type': 'application/json'})]
        mock_method, result = self._notify(response_list)
        self.assertFalse(result)

        self._validate_post_args(mock_method.call_args_list[0][1], 'json')
        self._validate_post_args(mock_method.call_args_list[1][1], 'data')

    def test_slack_post_message_success_cache_full(self):
        """url in cache and data is sent as raw data at first time
        """
        dummy_cache = [d for d in range(0, 100)]
        with mock.patch.object(slack_notifier.SlackNotifier,
                               '_raw_data_url_caches',
                               dummy_cache):
            response_list = [RequestsResponse(200, '{"ok":false,"error":"failure"}',
                                              {'Content-Type': 'application/json'}),
                             RequestsResponse(200, '{"ok":true}',
                                              {'Content-Type': 'application/json'})]
            mock_method, result = self._notify(response_list)
            self.assertTrue(result)
            self._validate_post_args(mock_method.call_args_list[0][1], 'json')
            self._validate_post_args(mock_method.call_args_list[1][1], 'data')
            self.assertEqual(dummy_cache,
                             slack_notifier.SlackNotifier._raw_data_url_caches)

    def test_config_insecure_true_ca_certs(self):
        self.conf_override(group='slack_notifier', timeout=50,
                           insecure=True,
                           ca_certs='/etc/ssl/certs/ca-bundle.crt')

        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]

        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        self.assertEqual(slack_notifier.SlackNotifier._raw_data_url_caches, [])
        post_args = mock_method.call_args_list[0][1]
        self.assertEqual(slack_text(),
                         json.loads(post_args.get('json').get('text')))
        self.assertEqual(50, post_args.get('timeout'))
        self.assertEqual('http://test.slack:3333', post_args.get('url'))
        self.assertEqual('/etc/ssl/certs/ca-bundle.crt', post_args.get('verify'))

    def test_config_insecure_true_no_ca_certs(self):
        self.conf_override(group='slack_notifier', timeout=50,
                           insecure=True,
                           ca_certs=None)

        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]

        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        self.assertEqual(slack_notifier.SlackNotifier._raw_data_url_caches, [])
        post_args = mock_method.call_args_list[0][1]
        self.assertEqual(slack_text(),
                         json.loads(post_args.get('json').get('text')))
        self.assertEqual(50, post_args.get('timeout'))
        self.assertEqual('http://test.slack:3333', post_args.get('url'))
        self.assertFalse(post_args.get('verify'))

    def test_config_insecure_false_no_ca_certs(self):
        self.conf_override(group='slack_notifier', timeout=50,
                           insecure=False,
                           ca_certs=None)

        response_list = [RequestsResponse(200, 'ok',
                                          {'Content-Type': 'application/text'})]

        mock_method, result = self._notify(response_list)
        self.assertTrue(result)
        mock_method.assert_called_once()
        self.assertEqual(slack_notifier.SlackNotifier._raw_data_url_caches, [])
        post_args = mock_method.call_args_list[0][1]
        self.assertEqual(slack_text(),
                         json.loads(post_args.get('json').get('text')))
        self.assertEqual(50, post_args.get('timeout'))
        self.assertEqual('http://test.slack:3333', post_args.get('url'))
        self.assertTrue(post_args.get('verify'))



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import simplejson as json

from debtcollector import removals
from oslo_config import cfg

from monasca_notification.plugins import abstract_notifier


CONF = cfg.CONF
VALID_HTTP_CODES = [200, 201, 204]


class PagerdutyNotifier(abstract_notifier.AbstractNotifier):

    type = 'pagerduty'

    def __init__(self, log):
        super(PagerdutyNotifier, self).__init__()
        self._log = log

    @removals.remove(
        message='Configuration of notifier is available through oslo.cfg',
        version='1.9.0',
        removal_version='3.0.0'
    )
    def config(self, config):
        pass

    @property
    def statsd_name(self):
        return 'sent_pagerduty_count'

    def send_notification(self, notification):
        """Send pagerduty notification
        """

        url = CONF.pagerduty_notifier.url
        headers = {"content-type": "application/json"}
        body = {"service_key": notification.address,
                "event_type": "trigger",
                "description": notification.message,
                "client": "Monasca",
                "client_url": "",
                "details": {"alarm_id": notification.alarm_id,
                            "alarm_name": notification.alarm_name,
                            "current": notification.state,
                            "message": notification.message}}

        try:
            result = requests.post(url=url,
                                   data=json.dumps(body),
                                   headers=headers,
                                   timeout=CONF.pagerduty_notifier.timeout)

            if result.status_code in VALID_HTTP_CODES:
                return True

            self._log.error("Error with pagerduty request. key=<{}> response={}"
                            .format(notification.address, result.status_code))
            return False
        except Exception:
            self._log.exception("Exception on pagerduty request. key=<{}>"
                                .format(notification.address))
            return False


pagerduty_notifier_group = cfg.OptGroup(name='%s_notifier' % PagerdutyNotifier.type)
pagerduty_notifier_opts = [
    cfg.IntOpt(name='timeout', default=5, min=1),
    cfg.StrOpt(name='url',
               default='https://events.pagerduty.com/'
                       'generic/2010-04-15/create_event.json')
]


def register_opts(conf):
    conf.register_group(pagerduty_notifier_group)
    conf.register_opts(pagerduty_notifier_opts, group=pagerduty_notifier_group)


def list_opts():
    return {
        pagerduty_notifier_group: pagerduty_notifier_opts
    }



# (C) Copyright 2016-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib

from debtcollector import removals
import jinja2
from oslo_config import cfg
import requests
import simplejson as json

from monasca_notification.plugins import abstract_notifier

CONF = cfg.CONF


class SlackNotifier(abstract_notifier.AbstractNotifier):
    """This module is a notification plugin to integrate with Slack.

    This plugin supports 2 types of APIs below.

    1st: Slack API
        notification.address = https://slack.com/api/chat.postMessage?token={token}&channel=#foobar

        You need to specify your token and channel name in the address.
        Regarding {token}, login to your slack account via browser and check the following page.
            https://api.slack.com/docs/oauth-test-tokens

    2nd: Incoming webhook
        notification.address = https://hooks.slack.com/services/foo/bar/buz

        You need to get the Incoming webhook URL.
        Login to your slack account via browser and check the following page.
            https://my.slack.com/services/new/incoming-webhook/
        Slack document about incoming webhook:
            https://api.slack.com/incoming-webhooks
    """

    type = 'slack'

    MAX_CACHE_SIZE = 100
    RESPONSE_OK = 'ok'

    _raw_data_url_caches = []

    def __init__(self, log):
        super(SlackNotifier, self).__init__()
        self._log = log

    @removals.remove(
        message='Configuration of notifier is available through oslo.cfg',
        version='1.9.0',
        removal_version='3.0.0'
    )
    def config(self, config_dict):
        pass

    @property
    def statsd_name(self):
        return 'sent_slack_count'

    def _build_slack_message(self, notification):
        """Builds slack message body
        """
        body = {'alarm_id': notification.alarm_id,
                'alarm_definition_id': notification.raw_alarm['alarmDefinitionId'],
                'alarm_name': notification.alarm_name,
                'alarm_description': notification.raw_alarm['alarmDescription'],
                'alarm_timestamp': notification.alarm_timestamp,
                'state': notification.state,
                'old_state': notification.raw_alarm['oldState'],
                'message': notification.message,
                'tenant_id': notification.tenant_id,
                'metrics': notification.metrics}

        slack_request = {}
        if CONF.slack_notifier.message_template:
            slack_request['text'] = self._render_message_template(body)
        else:
            slack_request['text'] = json.dumps(body, indent=3)

        return slack_request

    def _render_message_template(self, params):
        path, name = os.path.split(CONF.slack_notifier.message_template)
        loader = jinja2.FileSystemLoader(path)
        env = jinja2.Environment(loader=loader, autoescape=True)
        return env.get_template(name).render(params)

    def _check_response(self, result):
        if 'application/json' in result.headers.get('Content-Type'):
            response = result.json()
            if response.get(self.RESPONSE_OK):
                return True
            else:
                self._log.error('Received an error message when trying to send to slack. error={}'
                                .format(response.get('error')))
                return False
        elif self.RESPONSE_OK == result.text:
            return True
        else:
            self._log.error('Received an error message when trying to send to slack. error={}'
                            .format(result.text))
            return False

    def _send_message(self, request_options):
        try:
            url = request_options.get('url')
            result = requests.post(**request_options)
            if result.status_code not in range(200, 300):
                self._log.error('Received an HTTP code {} when trying to post on URL {}.'
                                .format(result.status_code, url))
                return False

            # Slack returns 200 ok even if the token is invalid. Response has valid error message
            if self._check_response(result):
                self._log.info('Notification successfully posted.')
                return True

            self._log.error('Failed to send to slack on URL {}.'.format(url))
            return False
        except Exception as err:
            self._log.exception('Error trying to send to slack on URL {}. Detail: {}'
                                .format(url, err))
            return False

    def send_notification(self, notification):
        """Send the notification via slack
            Posts on the given url
        """

        slack_message = self._build_slack_message(notification)

        address = notification.address
        #  '#' is reserved character and replace it with ascii equivalent
        #  Slack room has '#' as first character
        address = address.replace('#', '%23')

        parsed_url = urllib.parse.urlsplit(address)
        query_params = urllib.parse.parse_qs(parsed_url.query)
        # URL without query params
        url = urllib.parse.urljoin(address, urllib.parse.urlparse(address).path)

        # Default option is to do cert verification
        # If ca_certs is specified, do cert validation and ignore insecure flag
        verify = CONF.slack_notifier.ca_certs or not CONF.slack_notifier.insecure

        proxy = CONF.slack_notifier.proxy
        proxy_dict = None
        if proxy is not None:
            proxy_dict = {'https': proxy}

        data_format_list = ['json', 'data']
        if url in SlackNotifier._raw_data_url_caches:
            data_format_list = ['data']

        for data_format in data_format_list:
            self._log.info('Trying to send message to {} as {}'
                           .format(url, data_format))
            request_options = {
                'url': url,
                'verify': verify,
                'params': query_params,
                'proxies': proxy_dict,
                'timeout': CONF.slack_notifier.timeout,
                data_format: slack_message
            }
            if self._send_message(request_options):
                if (data_format == 'data' and
                        url not in SlackNotifier._raw_data_url_caches and
                        len(SlackNotifier._raw_data_url_caches) < self.MAX_CACHE_SIZE):
                    # NOTE:
                    #    There are a few URLs which can accept only raw data, so
                    #    only the URLs with raw data are kept in the cache. When
                    #    too many URLs exists, it can be considered malicious
                    #    user registers them.
                    #    In this case, older ones should be safer than newer
                    #    ones. When exceeding the cache size, do not replace the
                    #    the old cache with the newer one.
                    SlackNotifier._raw_data_url_caches.append(url)
                return True

            self._log.info('Failed to send message to {} as {}'
                           .format(url, data_format))
        return False


slack_notifier_group = cfg.OptGroup(name='%s_notifier' % SlackNotifier.type)
slack_notifier_opts = [
    cfg.IntOpt(name='timeout', default=5, min=1),
    cfg.BoolOpt(name='insecure', default=True),
    cfg.StrOpt(name='ca_certs', default=None),
    cfg.StrOpt(name='proxy', default=None),
    cfg.StrOpt(name='message_template', default=None)
]


def register_opts(conf):
    conf.register_group(slack_notifier_group)
    conf.register_opts(slack_notifier_opts, group=slack_notifier_group)


def list_opts():
    return {
        slack_notifier_group: slack_notifier_opts
    }



# -*- coding: utf-8 -*-

# Copyright 2016 Fujitsu Technology Solutions
# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from sqlalchemy import Column, String, Integer, Enum, DateTime, ForeignKey, Table

ALARM_STATES = ('UNDETERMINED', 'OK', 'ALARM')


def create_alarm_action_model(metadata=None):
    return Table('alarm_action', metadata,
                 Column('action_id',
                        String(36), ForeignKey('notification_method.id'),
                        nullable=False, primary_key=True),
                 Column('alarm_definition_id', String(36), primary_key=True),
                 Column('alarm_state', Enum(*ALARM_STATES), primary_key=True))


def create_notification_method_model(metadata=None):
    return Table('notification_method', metadata,
                 Column('id', String(36), primary_key=True),
                 Column('address', String(100)),
                 Column('name', String(250)),
                 Column('tenant_id', String(36)),
                 Column('type', String(255)),
                 Column('period', Integer),
                 Column('created_at', DateTime, default=lambda: datetime.utcnow()),
                 Column('updated_at', DateTime, onupdate=lambda: datetime.utcnow()))


def create_notification_method_type_model(metadata=None):
    return Table('notification_method_type', metadata,
                 Column('name', String(20), primary_key=True))


def create_alarm_model(metadata=None):
    return Table('alarm', metadata,
                 Column('id', String(20), primary_key=True),
                 Column('alarm_definition_id', String(36)),
                 Column('state', Enum(*ALARM_STATES)),
                 Column('lifecycle_state', String(50)),
                 Column('link', String(512)),
                 Column('created_at', DateTime, default=lambda: datetime.utcnow()),
                 Column('updated_at', DateTime, onupdate=lambda: datetime.utcnow()),
                 Column('state_updated_at', DateTime))



# (C) Copyright 2015-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import time
from unittest import mock

from monasca_notification import notification as m_notification
from monasca_notification.plugins import email_notifier
from monasca_notification.plugins import pagerduty_notifier
from monasca_notification.plugins import webhook_notifier
from monasca_notification.types import notifiers
from tests import base


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "LOW",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time(),
            "metrics": metrics}


class NotifyStub(object):
    def __init__(self, trap, config, send, failure):
        self.config_exception = config
        self.send_exception = send
        self.failure = failure
        self.trap = trap

    @property
    def type(self):
        return "email"

    @property
    def statsd_name(self):
        return "smtp_sent"

    def config(self, config_dict):
        if self.config_exception:
            raise Exception
        else:
            pass

    def send_notification(self, notification_obj):
        if self.send_exception:
            raise Exception
        else:
            if self.failure:
                return False
            else:
                return True


class Statsd(object):
    def __init__(self):
        self.timer = StatsdTimer()
        self.counter = StatsdCounter()

    def get_timer(self):
        return self.timer

    def get_counter(self, key):
        return self.counter


class StatsdTimer(object):
    def __init__(self):
        self.timer_calls = {}

    @contextlib.contextmanager
    def time(self, key):
        self.start(key)
        yield
        self.stop(key)

    def start(self, key):
        key = key + "_start"
        if key in self.timer_calls:
            self.timer_calls[key] += 1
        else:
            self.timer_calls[key] = 1

    def stop(self, key):
        key = key + "_stop"
        if key in self.timer_calls:
            self.timer_calls[key] += 1
        else:
            self.timer_calls[key] = 1


class StatsdCounter(object):
    def __init__(self):
        self.counter = 0

    def increment(self, val):
        self.counter += val


class TestInterface(base.BaseTestCase):

    def setUp(self):
        super(TestInterface, self).setUp()
        self.trap = []
        self.statsd = Statsd()

        email_notifier.register_opts(base.config.CONF)
        webhook_notifier.register_opts(base.config.CONF)
        pagerduty_notifier.register_opts(base.config.CONF)

        self.conf_override(group='email_notifier', server='my.smtp.server',
                           port=25, user=None, password=None,
                           timeout=60, from_addr='hpcs.mon@hp.com')

    def tearDown(self):
        super(TestInterface, self).tearDown()
        notifiers.possible_notifiers = []
        notifiers.configured_notifiers = {}
        self.trap = []

    def _configExceptionStub(self, log):
        return NotifyStub(self.trap, True, False, False)

    def _sendExceptionStub(self, log):
        return NotifyStub(self.trap, False, True, False)

    def _sendFailureStub(self, log):
        return NotifyStub(self.trap, False, False, True)

    def _goodSendStub(self, log):
        return NotifyStub(self.trap, False, False, False)

    def test_enabled_notifications_none(self):
        self.conf_override(
            group='notification_types',
            enabled=[]
        )

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = notifiers.enabled_notifications()

        self.assertEqual(len(notifications), 0)

    def test_enabled_notifications(self):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.email_notifier:EmailNotifier',
                'monasca_notification.plugins.pagerduty_notifier:PagerdutyNotifier',
                'monasca_notification.plugins.webhook_notifier:WebhookNotifier',
                'monasca_notification.plugins.hipchat_notifier:HipChatNotifier',
                'monasca_notification.plugins.slack_notifier:SlackNotifier',
                'monasca_notification.plugins.jira_notifier:JiraNotifier'
            ]
        )
        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = notifiers.enabled_notifications()
        self.assertEqual(len(notifications), 6)
        self.assertItemsEqual(notifications,
                              ['EMAIL', 'PAGERDUTY', 'WEBHOOK',
                               'HIPCHAT', 'SLACK', 'JIRA'])

    @mock.patch('monasca_notification.plugins.email_notifier')
    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    @mock.patch('monasca_notification.types.notifiers.importutils')
    def test_send_notification_exception(self, mock_im, mock_log,
                                         mock_smtp, mock_email):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.email_notifier:EmailNotifier'
            ]
        )

        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        mock_email.EmailNotifier = self._sendExceptionStub
        mock_im.import_class.return_value = mock_email.EmailNotifier

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = [
            m_notification.Notification(0, 'email', 'email notification',
                                        'me@here.com', 0, 0, alarm({}))
        ]

        notifiers.send_notifications(notifications)

        self.assertIn("send_notification exception for email", self.trap)

    @mock.patch('monasca_notification.plugins.email_notifier')
    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    @mock.patch('monasca_notification.types.notifiers.importutils')
    def test_send_notification_failure(self, mock_im, mock_log,
                                       mock_smtp, mock_email):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.email_notifier:EmailNotifier'
            ]
        )

        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        mock_email.EmailNotifier = self._sendFailureStub
        mock_im.import_class.return_value = mock_email.EmailNotifier

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = [
            m_notification.Notification(0, 'email', 'email notification',
                                        'me@here.com', 0, 0, alarm({}))
        ]

        sent, failed, invalid = notifiers.send_notifications(notifications)

        self.assertEqual(sent, [])
        self.assertEqual(len(failed), 1)
        self.assertEqual(invalid, [])

    @mock.patch('monasca_notification.types.notifiers.time')
    @mock.patch('monasca_notification.plugins.email_notifier')
    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    @mock.patch('monasca_notification.types.notifiers.importutils')
    def test_send_notification_correct(self, mock_im, mock_log, mock_smtp,
                                       mock_email, mock_time):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.email_notifier:EmailNotifier'
            ]
        )

        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append

        mock_email.EmailNotifier = self._goodSendStub
        mock_time.time.return_value = 42
        mock_im.import_class.return_value = mock_email.EmailNotifier

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = [
            m_notification.Notification(0, 'email', 'email notification',
                                        'me@here.com', 0, 0, alarm({})),
            m_notification.Notification(1, 'email', 'email notification',
                                        'foo@here.com', 0, 0, alarm({})),
            m_notification.Notification(2, 'email', 'email notification',
                                        'bar@here.com', 0, 0, alarm({}))
        ]

        sent, failed, invalid = notifiers.send_notifications(notifications)

        self.assertEqual(len(sent), 3)
        self.assertEqual(failed, [])
        self.assertEqual(invalid, [])

        for n in sent:
            self.assertEqual(n.notification_timestamp, 42)

    @mock.patch('monasca_notification.plugins.email_notifier')
    @mock.patch('monasca_notification.plugins.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    @mock.patch('monasca_notification.types.notifiers.importutils')
    def test_statsd(self, mock_im, mock_log, mock_smtp, mock_email):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.email_notifier:EmailNotifier'
            ]
        )

        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append

        mock_email.EmailNotifier = self._goodSendStub
        mock_im.import_class.return_value = mock_email.EmailNotifier

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        notifications = [
            m_notification.Notification(0, 'email', 'email notification',
                                        'me@here.com', 0, 0, alarm({})),
            m_notification.Notification(1, 'email', 'email notification',
                                        'foo@here.com', 0, 0, alarm({})),
            m_notification.Notification(2, 'email', 'email notification',
                                        'bar@here.com', 0, 0, alarm({}))
        ]

        notifiers.send_notifications(notifications)

        self.assertEqual(self.statsd.timer.timer_calls['email_time_start'], 3)
        self.assertEqual(self.statsd.timer.timer_calls['email_time_stop'], 3)
        self.assertEqual(self.statsd.counter.counter, 3)

    def test_plugin_load(self):
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.hipchat_notifier:HipChatNotifier',
                'monasca_notification.plugins.slack_notifier:SlackNotifier'
            ]
        )

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        notifiers.config()

        self.assertEqual(len(notifiers.possible_notifiers), 2)
        self.assertItemsEqual(['hipchat', 'slack'],
                              notifiers.configured_notifiers)

    @mock.patch('monasca_notification.types.notifiers.log')
    def test_invalid_plugin_load_exception_ignored(self, mock_log):
        mock_log.exception = self.trap.append
        self.conf_override(
            group='notification_types',
            enabled=[
                'monasca_notification.plugins.hipchat_notifier:UnknownPlugin',
                'monasca_notification.plugins.slack_notifier:SlackNotifier'
            ]
        )

        notifiers.init(self.statsd)
        notifiers.load_plugins()
        self.assertEqual(len(notifiers.possible_notifiers), 1)
        self.assertEqual(len(self.trap), 1)

        configured_plugins = ["email", "webhook", "pagerduty", "slack"]
        for plugin in notifiers.configured_notifiers:
            self.asssertIn(plugin.type in configured_plugins)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from oslo_config import cfg
from oslo_log import log as logging

from monasca_common.kafka import client_factory
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.processors import alarm_processor as ap
from monasca_notification.processors import notification_processor as np

log = logging.getLogger(__name__)
CONF = cfg.CONF


class NotificationEngine(object):
    def __init__(self):
        self._statsd = get_statsd_client()
        self._consumer = client_factory.get_kafka_consumer(
            CONF.kafka.url,
            CONF.kafka.group,
            CONF.kafka.alarm_topic,
            CONF.zookeeper.url,
            CONF.zookeeper.notification_path,
            CONF.kafka.legacy_kafka_client_enabled)
        self._producer = client_factory.get_kafka_producer(
            CONF.kafka.url,
            CONF.kafka.legacy_kafka_client_enabled)
        self._alarms = ap.AlarmProcessor()
        self._notifier = np.NotificationProcessor()

    def _add_periodic_notifications(self, notifications):
        for notification in notifications:
            period = str(notification.period)
            if period in CONF.kafka.periodic.keys():
                notification.notification_timestamp = time.time()
                self._producer.publish(CONF.kafka.periodic[period],
                                       [notification.to_json()])

    def run(self):
        finished_count = self._statsd.get_counter(name='alarms_finished_count')
        for alarm in self._consumer:
            log.debug('Received alarm >|%s|<', str(alarm))
            notifications, partition, offset = self._alarms.to_notification(alarm)
            if notifications:
                self._add_periodic_notifications(notifications)

                sent, failed = self._notifier.send(notifications)
                self._producer.publish(CONF.kafka.notification_topic,
                                       [i.to_json() for i in sent])
                self._producer.publish(CONF.kafka.notification_retry_topic,
                                       [i.to_json() for i in failed])

            self._consumer.commit()
            finished_count.increment()



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Copyright 2016 FUJITSU LIMITED

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import monascastatsd

from keystoneauth1 import exceptions as kaexception
from keystoneauth1 import loading as kaloading
from oslo_config import cfg
from oslo_log import log

from monasca_notification.common.repositories import exceptions
from monasca_notification.notification import Notification

LOG = log.getLogger(__name__)
CONF = cfg.CONF

NOTIFICATION_DIMENSIONS = {'service': 'monitoring',
                           'component': 'monasca-notification'}


def get_db_repo():
    repo_driver = CONF.database.repo_driver
    LOG.debug('Enabling the %s RDB repository', repo_driver)
    return repo_driver(CONF)


def construct_notification_object(db_repo, notification_json):
    try:
        notification = Notification(notification_json['id'],
                                    notification_json['type'],
                                    notification_json['name'],
                                    notification_json['address'],
                                    notification_json['period'],
                                    notification_json['retry_count'],
                                    notification_json['raw_alarm'])
        # Grab notification method from database to see if it was changed
        stored_notification = grab_stored_notification_method(db_repo, notification.id)
        # Notification method was deleted
        if stored_notification is None:
            LOG.debug("Notification method {0} was deleted from database. "
                      "Will stop sending.".format(notification.id))
            return None
        # Update notification method with most up to date values
        else:
            notification.name = stored_notification[0]
            notification.type = stored_notification[1]
            notification.address = stored_notification[2]
            notification.period = stored_notification[3]
            return notification
    except exceptions.DatabaseException:
        LOG.warn("Error querying mysql for notification method. "
                 "Using currently cached method.")
        return notification
    except Exception as e:
        LOG.warn("Error when attempting to construct notification {0}".format(e))
        return None


def grab_stored_notification_method(db_repo, notification_id):
    try:
        stored_notification = db_repo.get_notification(notification_id)
    except exceptions.DatabaseException:
        LOG.debug('Database Error.  Attempting reconnect')
        stored_notification = db_repo.get_notification(notification_id)

    return stored_notification


def get_statsd_client(dimensions=None):
    local_dims = dimensions.copy() if dimensions else {}
    local_dims.update(NOTIFICATION_DIMENSIONS)
    if CONF.statsd.enable:
        LOG.debug("Establishing connection with statsd on {0}:{1}"
                  .format(CONF.statsd.host, CONF.statsd.port))
        client = monascastatsd.Client(name='monasca',
                                      host=CONF.statsd.host,
                                      port=CONF.statsd.port,
                                      dimensions=local_dims)
    else:
        LOG.warn("StatsD monitoring disabled. Overriding monascastatsd.Client to use it offline")
        client = OfflineClient(name='monasca',
                               host=CONF.statsd.host,
                               port=CONF.statsd.port,
                               dimensions=local_dims)
    return client


class OfflineClient(monascastatsd.Client):

    def _set_connection(self, connection, host, port):
        if connection is None:
            self.connection = OfflineConnection(host=host,
                                                port=port,
                                                max_buffer_size=self._max_buffer_size)
        else:
            self.connection = connection


class OfflineConnection(monascastatsd.Connection):

    def __init__(self, host='localhost', port=8125, max_buffer_size=50):
        """Initialize an Offline Connection object.

        >>> monascastatsd = MonascaStatsd()

        :name: the name for this client.  Everything sent by this client
            will be prefixed by name
        :param host: the host of the MonascaStatsd server.
        :param port: the port of the MonascaStatsd server.
        :param max_buffer_size: Maximum number of metric to buffer before
         sending to the server if sending metrics in batch
        """
        self.max_buffer_size = max_buffer_size
        self._send = self._send_to_server
        self.connect(host, port)
        self.encoding = 'utf-8'

    def connect(self, host, port):
        """Avoid connecting to the monascastatsd server.

        """
        pass

    def _send_to_server(self, packet):
        pass


def get_keystone_session():

    auth_details = {}
    auth_details['auth_url'] = CONF.keystone.auth_url
    auth_details['username'] = CONF.keystone.username
    auth_details['password'] = CONF.keystone.password
    auth_details['project_name'] = CONF.keystone.project_name
    auth_details['user_domain_name'] = CONF.keystone.user_domain_name
    auth_details['project_domain_name'] = CONF.keystone.project_domain_name
    loader = kaloading.get_plugin_loader('password')
    auth_plugin = loader.load_from_options(**auth_details)
    session = kaloading.session.Session().load_from_options(
        auth=auth_plugin)
    return session


def get_auth_token():
    error_message = 'Keystone request failed: {}'
    try:
        session = get_keystone_session()
        auth_token = session.get_token()
        return auth_token
    except (kaexception.Unauthorized, kaexception.DiscoveryFailure) as e:
        LOG.exception(error_message.format(str(e)))
        raise
    except Exception as e:
        LOG.exception(error_message.format(str(e)))
        raise
